@@ -160,6 +160,9 @@ static unsigned canary_refusal_sec = 5*60;
 /* Demote root processes? */
 static bool canary_demote_root = FALSE;
 
+/* Demote unknown processes? */
+static bool canary_demote_unknown = FALSE;
+
 /* Log to stderr? */
 static bool log_stderr = FALSE;
 
@@ -874,6 +877,9 @@ static void reset_known(void) {
         struct rtkit_user *u;
         struct process *p;
         struct thread *t;
+        unsigned n_demoted = 0;
+
+        syslog(LOG_INFO, "Demoting known real-time threads.\n");
 
         for (u = users; u; u = u->next)
                 for (p = u->processes; p; p = p->next)
@@ -881,7 +887,16 @@ static void reset_known(void) {
                                 if (verify_process_user(u, p) >= 0 &&
                                     verify_process_starttime(p) >= 0 &&
                                     verify_thread_starttime(p, t) >= 0)
-                                        thread_reset(t->pid);
+                                        if (thread_reset(t->pid) >= 0) {
+                                                char exe[64];
+                                                syslog(LOG_NOTICE, "Successfully demoted thread %llu of process %llu (%s).\n",
+                                                       (unsigned long long) t->pid,
+                                                       (unsigned long long) p->pid,
+                                                       get_exe_name(p->pid, exe, sizeof(exe)));
+                                                n_demoted++;
+                                        }
+
+        syslog(LOG_NOTICE, "Demoted %u threads.\n", n_demoted);
 }
 
 static int reset_all(void) {
@@ -892,7 +907,7 @@ static int reset_all(void) {
         /* Goes through /proc and demotes *all* threads to
          * SCHED_OTHER */
 
-        syslog(LOG_INFO, "Rampaging.\n");
+        syslog(LOG_INFO, "Demoting known and unknown real-time threads.\n");
 
         if (!(pd = opendir(get_proc_path()))) {
                 r = -errno;
@@ -1540,10 +1555,14 @@ static void* watchdog_thread(void *data) {
 
                 if (TIMESPEC_MSEC(last_cheep) + canary_watchdog_msec <= TIMESPEC_MSEC(now)) {
                         last_cheep = now;
-                        syslog(LOG_WARNING, "The poor little canary died! Taking action.\n");
+                        syslog(LOG_WARNING, "The canary thread is apparently starving. Taking action.\n");
                         refuse_until = (uint32_t) now.tv_sec + canary_refusal_sec;
                         __sync_synchronize();
-                        reset_all();
+
+                        if (canary_demote_unknown)
+                                reset_all();
+                        else
+                                reset_known();
                         continue;
                 }
         }
@@ -1766,6 +1785,7 @@ enum {
         ARG_CANARY_CHEEP_MSEC,
         ARG_CANARY_WATCHDOG_MSEC,
         ARG_CANARY_DEMOTE_ROOT,
+        ARG_CANARY_DEMOTE_UNKNOWN,
         ARG_CANARY_REFUSE_SEC,
         ARG_STDERR,
         ARG_INTROSPECT
@@ -1794,6 +1814,7 @@ static const struct option long_options[] = {
     { "canary-cheep-msec",           required_argument, 0, ARG_CANARY_CHEEP_MSEC },
     { "canary-watchdog-msec",        required_argument, 0, ARG_CANARY_WATCHDOG_MSEC },
     { "canary-demote-root",          no_argument,       0, ARG_CANARY_DEMOTE_ROOT },
+    { "canary-demote-unknown",       no_argument,       0, ARG_CANARY_DEMOTE_UNKNOWN },
     { "canary-refuse-sec",           required_argument, 0, ARG_CANARY_REFUSE_SEC },
     { "stderr",                      no_argument,       0, ARG_STDERR },
     { "introspect",                  no_argument,       0, ARG_INTROSPECT },
@@ -1842,6 +1863,8 @@ static void show_help(const char *exe) {
                "      --actions-per-burst-max=INT     Allow this many requests per burst (%u)\n\n"
                "      --canary-cheep-msec=MSEC        Canary cheep interval (%u)\n"
                "      --canary-watchdog-msec=MSEC     Watchdog action delay (%u)\n"
+               "      --canary-demote-unknown         When the canary dies demote unknown\n"
+               "                                      processes too?\n"
                "      --canary-demote-root            When the canary dies demote root\n"
                "                                      processes too?\n"
                "      --canary-refuse-sec=SEC         How long to refuse further requests\n"
@@ -2089,6 +2112,10 @@ static int parse_command_line(int argc, char *argv[], int *ret) {
                                 canary_demote_root = TRUE;
                                 break;
 
+                        case ARG_CANARY_DEMOTE_UNKNOWN:
+                                canary_demote_unknown = TRUE;
+                                break;
+
                         case ARG_CANARY_REFUSE_SEC: {
                                 char *e = NULL;
                                 unsigned long u;
